// -*- C++ -*-
/*!
 * @file  BCN3D_ArmRTC.cpp
 * @brief BCN3D_ArmRTC
 * @date $Date$
 *
 * $Id$
 */

#include "BCN3D_ArmRTC.h"

#define DEGREE_TO_RADIAN(x)		((M_PI/180.0) * (x))

using namespace std;
using namespace cnoid;

namespace{
	const double timeStep = 0.001;
};

// Module specification
// <rtc-template block="module_spec">
static const char* bcn3d_armrtc_spec[] =
  {
    "implementation_id", "BCN3D_ArmRTC",
    "type_name",         "BCN3D_ArmRTC",
    "description",       "BCN3D_ArmRTC",
    "version",           "1.0.0",
    "vendor",            "FSK",
    "category",          "Category",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    ""
  };

//}
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
BCN3D_ArmRTC::BCN3D_ArmRTC(RTC::Manager* manager)
    // <rtc-template block="initializer">
//  : RTC::DataFlowComponentBase(manager),
  : BodyIoRTC(manager),
	m_torquesIn("u", m_torques),
    m_anglesOut("q", m_angles)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
BCN3D_ArmRTC::~BCN3D_ArmRTC()
{
}



/*RTC::ReturnCode_t BCN3D_ArmRTC::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("u", m_torquesIn);
  
  // Set OutPort buffer
  addOutPort("q", m_anglesOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // </rtc-template>
  
  return RTC::RTC_OK;
}*/

bool BCN3D_ArmRTC::initializeIO(ControllerIO* io)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("u", m_torquesIn);

  // Set OutPort buffer
  addOutPort("q", m_anglesOut);

  // Set service provider to Ports

  // Set service consumers to Ports

  // Set CORBA Service Ports

  // </rtc-template>

  // <rtc-template block="bind_config">
  // </rtc-template>
  m_angles.data.length(11);

  return true;
}

bool BCN3D_ArmRTC::initializeSimulation(ControllerIO* io)
{
	// Set the body.
	ioBody = io->body();

	// Set the links.
	link1M1		= ioBody->link("1M1");
	link2M2		= ioBody->link("2M2");
	link3M2		= ioBody->link("3M2");
	link4M2		= ioBody->link("4M2");
	link4M1		= ioBody->link("4M1");
	idolGear	= ioBody->link("Idol_Gear");
	gripperL	= ioBody->link("Gripper_Left");
	pivotArmL	= ioBody->link("Pivot_Arm_L");
	servoGear	= ioBody->link("Servo_Gear");
	gripperR	= ioBody->link("Gripper_Right");
	pivotArmR	= ioBody->link("Pivot_Arm_R");

	// Set the actuation mode.
	link1M1->setActuationMode(Link::JOINT_TORQUE);
	link2M2->setActuationMode(Link::JOINT_TORQUE);
	link3M2->setActuationMode(Link::JOINT_TORQUE);
	link4M2->setActuationMode(Link::JOINT_TORQUE);
	link4M1->setActuationMode(Link::JOINT_TORQUE);
	idolGear->setActuationMode(Link::JOINT_TORQUE);
	gripperL->setActuationMode(Link::JOINT_TORQUE);
	pivotArmL->setActuationMode(Link::JOINT_TORQUE);
	servoGear->setActuationMode(Link::JOINT_TORQUE);
	gripperR->setActuationMode(Link::JOINT_TORQUE);
	pivotArmR->setActuationMode(Link::JOINT_TORQUE);

	// Set the angles.
	m_angles.data[0]	= DEGREE_TO_RADIAN(0);			// 1M1
	m_angles.data[1]	= DEGREE_TO_RADIAN(-110);		// 2M2
	m_angles.data[2]	= DEGREE_TO_RADIAN(-120);		// 3M2
	m_angles.data[3]	= DEGREE_TO_RADIAN(110);		// 4M2
	m_angles.data[4]	= DEGREE_TO_RADIAN(0);			// 4M1
	m_angles.data[5]	= DEGREE_TO_RADIAN(0);			// Idol Gear
	m_angles.data[6]	= DEGREE_TO_RADIAN(0);			// Gripper_Left
	m_angles.data[7]	= DEGREE_TO_RADIAN(0);			// Pivot_Arm_L
	m_angles.data[8]	= DEGREE_TO_RADIAN(0);			// Servo Gear
	m_angles.data[9]	= DEGREE_TO_RADIAN(0);			// Gripper_Right
	m_angles.data[10]	= DEGREE_TO_RADIAN(0);			// Pivot_Arm_R

	if(m_angles.data.length() >= 11){
		// In case of length not less than 11.
		for(int i = 0; i < 11; ++i){
			qref[i] = qprev[i] = m_angles.data[i];
		}
	}
	return true;
}

void BCN3D_ArmRTC::inputFromSimulator()
{
	m_angles.data[0]	= link1M1->q();
	m_angles.data[1]	= link2M2->q();
	m_angles.data[2]	= link3M2->q();
	m_angles.data[3]	= link4M2->q();
	m_angles.data[4]	= link4M1->q();
	m_angles.data[5]	= idolGear->q();
	m_angles.data[6]	= gripperL->q();
	m_angles.data[7]	= pivotArmL->q();
	m_angles.data[8]	= servoGear->q();
	m_angles.data[9]	= gripperR->q();
	m_angles.data[10]	= pivotArmR->q();

	m_anglesOut.write();
}

void BCN3D_ArmRTC::outputToSimulator()
{
	double q[11] = { qprev[0], qprev[1], qprev[2], qprev[3], qprev[4], qprev[5],
			         qprev[6], qprev[7], qprev[8], qprev[9], qprev[10] };

	if(m_angles.data.length() >= 11){
		// In case of length not less than 11.
		for(int i = 0; i < 11; ++i){
			q[i] = m_angles.data[i];
		}
	}

	static const double P[11] = { 250, 250, 150, 150, 100,
				                  3, 3, 3, 3, 3, 3 };
	static const double D[11] = { 125, 125, 75, 75, 50,
								  0.3, 0.3, 0.3, 0.3, 0.3, 0.3 };

	if(m_torquesIn.isNew()){
		// When a new value of torque is input.

		// Reading the torques value.
		m_torquesIn.read();

		if(m_torques.data.length() >= 11){
			// In case of length not less than 11.
			for(int i = 0; i < 11; ++i){
				double dq = (q[i] - qprev[i]) / timeStep;
				double dqref = 0.0;
				double pos = m_torques.data[i];

				if(fabs(pos) > 0.25){
					double deltaq = 0.001 * pos / 4;
					qref[i] += deltaq;
					dqref = deltaq / timeStep;
				}

				m_torques.data[i] = P[i] * (qref[i] - q[i]) + D[i] * (dqref - dq);

				qprev[i] = q[i];

			}

			link1M1->u()	= m_torques.data[0];
			link2M2->u()	= m_torques.data[1];
			link3M2->u()	= m_torques.data[2];
			link4M2->u()	= m_torques.data[3];
			link4M1->u()	= m_torques.data[4];
			idolGear->u()	= m_torques.data[5];
			gripperL->u()	= m_torques.data[6];
			pivotArmL->u()	= m_torques.data[7];
			servoGear->u()	= m_torques.data[8];
			gripperR->u()	= m_torques.data[9];
			pivotArmR->u()	= m_torques.data[10];
		}
	}

}
/*
RTC::ReturnCode_t BCN3D_ArmRTC::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t BCN3D_ArmRTC::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t BCN3D_ArmRTC::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t BCN3D_ArmRTC::onActivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t BCN3D_ArmRTC::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t BCN3D_ArmRTC::onExecute(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t BCN3D_ArmRTC::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t BCN3D_ArmRTC::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t BCN3D_ArmRTC::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t BCN3D_ArmRTC::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t BCN3D_ArmRTC::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void BCN3D_ArmRTCInit(RTC::Manager* manager)
  {
    coil::Properties profile(bcn3d_armrtc_spec);
    manager->registerFactory(profile,
                             RTC::Create<BCN3D_ArmRTC>,
                             RTC::Delete<BCN3D_ArmRTC>);
  }
  
};


